!=========================================================================
! Copyright (C) GemTalk Systems 1986-2016.  All Rights Reserved.
!
! $Id: classhistory.gs,v 1.7 2008-01-09 19:02:30 stever Exp $
!
! Superclass Hierarchy:
!   ClassHistory, Array, SequenceableCollection, Collection, Object.
!
!=========================================================================

removeallmethods ClassHistory
removeallclassmethods ClassHistory

category: 'For Documentation Installation only'
classmethod: ClassHistory
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A ClassHistory is a sequence of Class objects that logically represent the
 historical revisions to a Class.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'A textual description of the function of the Class.'.
doc documentInstVar: #description with: txt.

txt := (GsDocText new) details:
'The class history''s name for itself; a Symbol of up to 64 Characters.'.
doc documentInstVar: #name with: txt.

self description: doc.
%

category: 'Instance Creation'
classmethod: ClassHistory
new

"Create a new ClassHistory."

^ super new
%

category: 'Accessing'
method: ClassHistory
at: aTimeOrIndex

"Returns the Class that was current at the given time.  The time may be
 specified absolutely using a DateTime, or relatively using an integer.  If a
 DateTime is specified, returns the version of the class that was active at
 that time, or nil if the time is before the earliest version.

 If an Integer is specified, it is used to chronologically select the
 version, with 1 indicating the first version created, 2 the version, and so on.
 If the index is less than one or greater than the number of versions in
 the history, an error is generated."

| candidate aClass|

(aTimeOrIndex class == DateTime)
  ifFalse: [ ^ super at: aTimeOrIndex ]
  ifTrue: [
    "candidate is initially nil.  Iterate through the history from
    the beginning.  If the entry's time is before the specified time,
    it is possibly the class that was active so save it as the best
    candidate.  If the entry's time is after the specified time, it
    cannot be the answer and no subsequent ones can be, so returns the
    current candidate."

    1 to: self size do: [ :j |
      aClass := self at: j.
      (aClass timeStamp <= aTimeOrIndex)
        ifTrue: [ candidate := aClass ]
        ifFalse: [ ^ candidate ].
    ].
    ^ candidate
  ]
%

category: 'Accessing'
method: ClassHistory
current

"Returns the current, or most recent class."

^ super at: self size
%

category: 'Accessing'
method: ClassHistory
description

"Returns the description of this ClassHistory."

^ description
%

category: 'Accessing'
method: ClassHistory
name

"Returns the name of this ClassHistory."

^ name
%

category: 'Updating'
method: ClassHistory
description: aString

"Updates the description of this ClassHistory."

description := aString
%

category: 'Updating'
method: ClassHistory
name: aString

"Updates the name of this ClassHistory."

name := aString asString asSymbol
%

category: 'Accessing'
method: ClassHistory
currentVersion

"Returns the most recent version in the receiver's collection of versions."

| sz |
(sz := self size) == 0 ifTrue: [ ^nil ].
^self at: sz
%

category: 'Updating'
method: ClassHistory
newVersion: aClass

"Installs the given class as the receiver's most current version.  Does not
 install the receiver in the given class as its version history.
 Returns the class object."

self add: aClass
%

category: 'Updating'
method: ClassHistory
removeVersion: aClass

"Removes the given class from the receiver's list of versions."

| idx |
idx := self indexOf: aClass.
idx > 0 ifTrue: [
  self removeFrom: idx to: idx
].
%

! Fix 41012 - delete unifyClassHistories:
